from urllib.parse import urlparse
from re import sub

import requests
from requests.exceptions import ConnectionError

from pylaunch.xmlparse import XMLFile, normalize


def _prepare_url(url):
    parsed = urlparse(url)
    if parsed.netloc == "":
        base, *path = parsed.path.split("/")
        url = (
            f"http://{base}:8060/{'/'.join(path)}"
            if ":" not in base
            else f"http://{base}/{'/'.join(path)}"
        )
        return url
    return f"http://{parsed.netloc}{parsed.path}"


class Controller:
    def __init__(self, address: str):
        self.bind(address)

    def bind(self, address: str) -> None:
        """
        A function called on __init__ to bind to a specific device.
        """
        url = _prepare_url(address)
        try:
            response = self.request.get(url)
            app_url = response.headers.get("Application-URL")

            self.address = _prepare_url(app_url) if app_url else url

            xml = XMLFile(response.text)
            for element in xml.find("device"):
                key, value = normalize(xml, element)
                setattr(self, key, value) if value else None
        except ConnectionError:
            self.address = url

    def __getitem__(self, prop):
        return self.__getattribute__(prop)

    def __enter__(self):
        self._session = requests.Session()
        return self

    def __exit__(self, *args):
        self._session.close()

    def __repr__(self):
        return f"{self.__class__.__name__}({self.address!r})"

    @property
    def request(self) -> requests:
        if "_session" in dir(self):
            return self._session
        return requests



from __future__ import annotations
from typing import List, Dict, Callable
from re import sub
from urllib.parse import unquote, urlencode

from requests import Response

from pylaunch.core import Controller
from pylaunch.ssdp import SimpleServiceDiscoveryProtocol, ST_DIAL
from pylaunch.xmlparse import XMLFile, normalize


class AppNotFoundError(Exception):
    pass


class Dial(Controller):
    @classmethod
    def discover(cls, timeout: int = 3) -> List[Dial]:
        """
        Scans network for dial compatible devices and returns a list.
        """
        results = []
        ssdp = SimpleServiceDiscoveryProtocol(ST_DIAL)
        ssdp.settimeout(timeout)
        response = ssdp.broadcast()
        for resp in response:
            location = resp.headers.get("location")
            if not location:
                continue
            results.append(cls(location))
        return results

    @classmethod
    def from_xml(cls, xml_text) -> Dial:
        xml = XMLFile(xml_text)
        for element in xml.find("device"):
            key, value = normalize(xml, element)
            setattr(cls, key, value) if value else None

    def _build_app_url(self, app_name: str = None) -> str:
        """
        Simple helper function to build app urls.
        """
        return "/".join([self.address, app_name])

    def launch_app(
        self, app_name: str, callback: Callable[[None], Response] = None, **kwargs
    ) -> None:
        """
        Launches specified application.
        """
        url = self._build_app_url(app_name)
        data = unquote(urlencode(kwargs))
        headers = (
            {"Content-Type": "text/plain; charset=utf-8"}
            if kwargs
            else {"Content-Length": "0"}
        )
        resp = self.request.post(url, data=data, headers=headers)

        if resp.status_code < 300:
            self.instance_url = resp.headers.get("location")
            self.refresh_url = unquote(resp.text)
            callback(resp) if callback else None
        elif resp.status_code == 404:
            raise AppNotFoundError(f"No application found with name {app_name}")
        else:
            resp.raise_for_status()

    def kill_app(
        self, app_name: str = None, callback: Callable[[None], Response] = None
    ) -> None:
        """
        This will kill any active application tracked by this 
        instance if one exists and will return True if successful 
        otherwise it will return False.
        """
        if app_name:
            app_url = self._build_app_url(app_name) + "/run"
        elif not app_name and not self.instance_url:
            raise Exception("There is no instance found to kill.")
        else:
            app_url = self.instance_url
        resp = self.request.delete(app_url)
        if resp.status_code in [200, 204]:
            self.instance_url = None
            self.refresh_url = None
        elif resp.status_code == 404:
            raise Exception(f"There is no running {app_name} instance.")
        callback(resp) if callback else None

    def get_app_status(self, app_name: str) -> Dict[str, str]:
        """
        Makes a request to the DIAL device with a application name parameter and returns
        a dictionary including the supported DIAL version, app name and state.
        State examples: running, stopped or installable
        """
        url = self._build_app_url(app_name)
        resp = self.request.get(url, headers={"Content-Type": "text/plain"})
        if resp.status_code == 404:
            raise AppNotFoundError(f"No application found with name {app_name}")
        xml = XMLFile(resp.text)
        return {
            "version": xml.find("service").attrib.get("dialVer"),
            "name": xml.find("name").text,
            "state": xml.find("state").text,
        }



from re import match, sub
from xml.etree import ElementTree


def normalize(xml, element):
    key = lambda element: element.tag.replace(xml.namespace, "")
    truther = lambda x: {"true": True, "false": False, "\n": None}.get(x, x)
    name = (
        "_".join(sub("([a-z])([A-Z])", r"\1 \2", key(element)).split())
        .lower()
        .replace("-", "_")
    )
    value = truther(element.text)
    return (name, value)


class XMLFile:
    def __init__(self, xml_string):
        self._root = ElementTree.fromstring(xml_string)

    def __getitem__(self, key):
        return self.__dict__[key]

    @property
    def namespace(self):
        m = match(r"\{.*\}", self._root.tag)
        return m.group(0) if m else ""

    @classmethod
    def from_file(cls, path):
        raise NotImplementedError

    def find(self, tag_name):
        """Recursively loops through every item in a tree to find the
        first element with the desired tag name."""
        tag = self.namespace + tag_name

        def find(element, tag):
            if element.tag != tag:
                for el in element:
                    element = find(el, tag)
                    if element.tag == tag:
                        return element
                    else:
                        continue
            return element

        return find(self._root, tag)
